//-----------------------------------------------------------------------------
// Copyright 2024 Peter Balch
// public domain
// magnetometer
//  
//  magnetometer 3 
//  ==============
//  
//  The circuit has magnetometer "transformer" 
//  The sketch sends approx 2mS pos/neg pulses to the input coil 
//  The output coil is connected to to the ADC
//    0xAA
//    0xBB
//    0xDD
//    positive threshold pulse (16-bit)
//    negative threshold pulse (16-bit)
//    conparator times  (4 of 16-bit)
//    timer: length of two drive pulses
//    ADCBUFFER: 8-bit values
//  Magnetometer.exe program on the PC displays the values
//  you decide what the threshold shold be
//-----------------------------------------------------------------------------

//#include <TimerOne.h>

//-----------------------------------------------------------------------------
// Defines and Typedefs
//-----------------------------------------------------------------------------

// get register bit - faster: doesn't turn it into 0/1
#ifndef getBit
#define getBit(sfr, bit) (_SFR_BYTE(sfr) & _BV(bit))
#endif

//-----------------------------------------------------------------------------
// Global Constants
//-----------------------------------------------------------------------------

const long BAUDRATE  = 115200;  // Baud rate of UART in bps

const int adjustPin = 8; // adjust threshold up/down
const int HbridgePinA = 2; // H-bridge pin A
const int HbridgePinB = 3; // H-bridge pin B
const int ComparatorNeg = 7; // Comparator input
const int ComparatorPos = 6; // Comparator input

//-----------------------------------------------------------------------------
// Global Variables
//-----------------------------------------------------------------------------

uint8_t curAdcPin = 0;
uint8_t curVref = 0;
uint8_t curADCPrescaler = 2;

const int ADCBUFFERSIZE = 1000;
uint8_t ADCBuffer[ADCBUFFERSIZE];
uint16_t times[4];

uint16_t tmr;
  
//-----------------------------------------------------------------------------
// initADC()
//-----------------------------------------------------------------------------
void initADC(void)
{
  //  ADCSRA = 0x00; // ADC Control and Status Register A
  // 0  Bit 7  ADEN: ADC Enable
  // 0  Bit 6  ADSC: ADC Start Conversion
  // 1  Bit 5  ADATE: ADC Auto Trigger Enable
  // 0  Bit 4  ADIF: ADC Interrupt Flag
  // 1  Bit 3  ADIE: ADC Interrupt Enable
  // 111  Bits 2:0  ADPSn: ADC Prescaler Select [n = 2:0]

  ADCSRB = 0x0 ; //ADC Control and Status Register B
  // 0 Bit 6  ACME: Analog Comparator Multiplexer Enable
  // 000 Bits 2:0  ADTSn: ADC Auto Trigger Source [n = 2:0] Free Running mode

  ADMUX = 0b01100000 + (curVref << 6) + curAdcPin; // ADC Multiplexer Selection Register
  // rr    Bits 7:6  REFSn: Reference Selection = Vcc
  // 1     Bit 5  ADLAR: ADC Left adjustPin Result
  // aaaa  Bits 3:0  MUXn: Analog Channel Selection

  DIDR0 = 0x3F; // Digital Input Disable Register 0
  // ADC0D=1, ADC1D=1, ADC2D=1, ADC3D=1, ADC4D=1, ADC5D=1, ADC6D=0, ADC7D=0

}

//-----------------------------------------------------------------------------
// GetADCSamples
// read and Tx a buffer-full of samples
// prescaler
//   7 128
//   6 64
//   5 32
//   4 16
//   3 8
//   2 4
//   1 2
//   0 2
//-----------------------------------------------------------------------------

void GetADCSamples(void) {
  uint8_t d,b;
  uint8_t *p;
  const int hysteresis = 2;
  enum TMode {waitNone,waitFalling,waitRising};
  TMode mode = waitNone;

  initADC();

  ADCSRA = 0x80 + (curADCPrescaler & 7); // ADC Control and Status Register A

  for (d = 0; d < 10; d++ ) { // make sure ADC is running
    bitSet(ADCSRA, ADSC); // start ADC conversion
    while (!getBit(ADCSRA, ADIF)) ; // wait for ADC
    bitSet(ADCSRA, ADIF); // clear the flag
  }

//  digitalWrite(HbridgePinA,HIGH);
//  digitalWrite(HbridgePinB,LOW);
  
  bitSet(ADCSRA, ADSC); // start ADC conversion
  for (p = ADCBuffer; p < ADCBuffer + ADCBUFFERSIZE/2; p++ ) {
//      bitSet(ADCSRA, ADSC); // start ADC conversion
    while (!getBit(ADCSRA, ADIF)) ; // wait for ADC
    bitSet(ADCSRA, ADIF); // clear the flag
    bitSet(ADCSRA, ADSC);
  }
  
  digitalWrite(adjustPin,HIGH);
  delayMicroseconds(5);
  TIFR1 |= (1<<ICF1); // clear interrupt flag by writing '1' 
  bitSet(TCCR1B,ICES1); // set comparator to trigger on rising edge
  mode = waitRising;
  TCNT1 = 0;
  digitalWrite(HbridgePinA,LOW);
  digitalWrite(HbridgePinB,HIGH);
  for (p = ADCBuffer; p < ADCBuffer + ADCBUFFERSIZE/2; p++ ) {
    while (!getBit(ADCSRA, ADIF)) ; // wait for ADC
    bitSet(ADCSRA, ADIF); // clear the ADC flag
    bitSet(ADCSRA, ADSC); // start ADC conversion
    b = ADCH;
    if (ACSR & (1<<ACO)) 
      bitSet(b,0); else
      bitClear(b,0); 
    *p = b;

    if (TIFR1 & (1<<ICF1)) // comparator triggered interrupt flag
    switch (mode) {
      case waitRising:
        times[0] = ICR1;
//times[0] = 1000;
        mode = waitFalling;
        bitClear(TCCR1B,ICES1); // set comparator to trigger on falling edge
        TIFR1 |= (1<<ICF1); // clear interrupt flag by writing '1' 
        break;

      case waitFalling:
        times[1] = ICR1;
//times[1] = 2000;
        mode = waitNone;
        break;
    }
  }
  
  digitalWrite(adjustPin,LOW);  
  delayMicroseconds(5);
  TIFR1 |= (1<<ICF1); // clear interrupt flag by writing '1' 
  bitClear(TCCR1B,ICES1); // set comparator to trigger on falling edge
  mode = waitFalling;
  
  digitalWrite(HbridgePinA,HIGH);
  digitalWrite(HbridgePinB,LOW);
  for (; p < ADCBuffer + ADCBUFFERSIZE; p++ ) {
    while (!getBit(ADCSRA, ADIF)) ; // wait for ADC
    bitSet(ADCSRA, ADIF); // clear the ADC flag
    bitSet(ADCSRA, ADSC); // start ADC conversion
    b = ADCH;
    if (ACSR & (1<<ACO)) 
      bitSet(b,0); else
      bitClear(b,0); 
    *p = b;

    if (TIFR1 & (1<<ICF1)) // comparator triggered interrupt flag
    switch (mode) {
      case waitFalling:
        times[2] = ICR1;
//times[2] = 3000;
        mode = waitRising;
        bitSet(TCCR1B,ICES1); // set comparator to trigger on rising edge
        TIFR1 |= (1<<ICF1); // clear interrupt flag by writing '1' 
        break;
        
      case waitRising:
        times[3] = ICR1;
//times[3] = 4000;
        mode = waitNone;
        break;
    }
  }

  digitalWrite(HbridgePinA,LOW);
  digitalWrite(HbridgePinB,LOW);
  tmr = TCNT1;
}

//-----------------------------------------------------------------------------
// SendADC
//-----------------------------------------------------------------------------

void SendADC(void) {
  int i;
  memset( (void *)ADCBuffer, 0, sizeof(ADCBuffer) );

  noInterrupts();

  initADC();
  GetADCSamples();
    
  interrupts();

//  Serial.println(times[0]-times[1]-times[2]+times[3]);

  Serial.write((int)0xAA);
  Serial.write((int)0xBB);
  Serial.write((int)0xDD);

//  analogReference(EXTERNAL);
  analogReference(DEFAULT);
  analogRead(A1);  
  digitalWrite(adjustPin,HIGH);
  delay(1);
  i = analogRead(A1);  
  Serial.write((uint8_t *)&i,2);
  digitalWrite(adjustPin,LOW);
  delay(1);
  i = analogRead(A1);  
  Serial.write((uint8_t *)&i,2);

  for (i=0; i<=3; i++)
    Serial.write((uint8_t *)&(times[i]),2);
  
  Serial.write((uint8_t *)&tmr,2);

  Serial.write( (uint8_t *)ADCBuffer, ADCBUFFERSIZE);
}

//-----------------------------------------------------------------------------
// Main routines
// The setup function
//-----------------------------------------------------------------------------
void setup (void) {    // Setup of the microcontroller
  Serial.begin(BAUDRATE);
  
  // Activate interrupts
  sei();

  initADC();

  Serial.println("Magnetometer ready");
  
  pinMode(adjustPin,OUTPUT);
  pinMode(ComparatorNeg,INPUT);
  pinMode(ComparatorPos,INPUT);
  pinMode(A0,INPUT);
  pinMode(A1,INPUT);

  pinMode(HbridgePinA,OUTPUT); 
  pinMode(HbridgePinB,OUTPUT); 

  curAdcPin = 0;
  curVref = 0;
  curADCPrescaler = 2;
  
  // timer1 
  TCCR1A = 0; // pwm off
  TCCR1B = (1<<ICNC1) // 0=disable Noise Canceler
         | (1<<ICES1) // Input Capture Edge Select 0 falling 1 rising
         | (0<<WGM12) | (0<<WGM13) // disable Waveform Generation 
         | 1; // prescaler

  // comparator
  ADCSRB = (0<<ACME); // comparator neg input from D7 (Multiplexer disable)
  ACSR = (0<<ACD) // turn on comparator 
       | (0<<ACBG) // comparator pos input from D6
       | (0<<ACO)  // Analog Comparator Output
       | (1<<ACI) // Analog Comparator Interrupt Flag; set by hardware; cleared by writing a logic one to the flag
       | (0<<ACIE) // Analog Comparator Interrupt Enable
       | (1<<ACIC) // Analog Comparator Input Capture Enable for Timer1
       | (1<<ACIS1) | (1<<ACIS0); // Interrupt Mode: 00 on Toggle; 10 on Falling Output Edge; 11 on Rising 
  
  TIMSK1 = 0; // Turn off timer1 interrupt-on-capture
}

void loop (void) {
  SendADC(); 
}
